import os
# Decorator to create directory and remove it after test

def set_and_clean_test_dir(dir_path: str):
    def inner_decorator_wo_arugments(func):
        def wrapper(*args, **kwargs):
            # Create directory
            os.makedirs(dir_path, exist_ok=True)
            # Run test
            func(*args, **kwargs)
            # Remove directory
            os.removedirs(dir_path)
        return wrapper
    return inner_decorator_wo_arugments


import os
import json
import string
import random
import unittest

import src.hkkang_utils.file as file_utils


class testing_environment:
    def __init__(self, dir_name=None, num_files=10, extension=".txt"):
        self.dir_name = dir_name if dir_name else self._generate_random_string()
        self.num_files = num_files
        self.extension = extension
        self.file_names = [
            self._generate_random_string() + extension for _ in range(num_files)
        ]

    @property
    def dummpy_dict_data(self):
        return {"key": "value"}

    @property
    def file_paths(self):
        return [os.path.join(self.dir_name, file_name) for file_name in self.file_names]

    # Helper functions
    def _generate_random_string(self, num_chars=10):
        letters = string.ascii_lowercase
        return "".join(random.choice(letters) for i in range(num_chars))

    # Handle directory
    def create_directory(self):
        os.mkdir(self.dir_name)

    def remove_directory(self):
        os.rmdir(self.dir_name)

    # Handle files
    def create_random_files(self):
        for file_name in self.file_names:
            file_path = os.path.join(self.dir_name, file_name)
            with open(file_path, "w") as f:
                json.dump(self.dummpy_dict_data, f)

    def remove_random_files(self):
        for file_name in self.file_names:
            file_path = os.path.join(self.dir_name, file_name)
            os.remove(file_path)

    # Magic methods
    def __enter__(self):
        self.create_directory()
        self.create_random_files()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.remove_random_files()
        self.remove_directory()


class Test_file_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_file_utils, self).__init__(*args, **kwargs)

    def test_get_files_in_directory(self):
        with testing_environment() as env:
            retrieved_file_paths = file_utils.get_files_in_directory(
                env.dir_name, return_with_dir=True
            )
            self.assertEqual(len(retrieved_file_paths), env.num_files)
            self.assertSetEqual(set(retrieved_file_paths), set(env.file_paths))

    def test_read_json_file(self):
        with testing_environment() as env:
            file_path = os.path.join(env.dir_name, env.file_names[0])
            json_data = file_utils.read_json_file(file_path)
            self.assertEqual(json_data, env.dummpy_dict_data)

    def test_get_files_in_all_sub_directories(self):
        pass

    def test_yaml_file_functions(self):
        yaml_path = "test.yaml"
        dict_object = {"A": "a", "B": {"C": "c", "D": "d", "E": "e"}}
        # Write yaml file
        file_utils.write_yaml_file(dict_object, yaml_path)

        # Read in yaml file
        yaml_object = file_utils.read_yaml_file(yaml_path)

        # Check if the two objects are the same
        self.assertEqual(dict_object, yaml_object)

        # Delete yaml file
        os.remove(yaml_path)

    def test_csv_file_functions(self):
        header = ["A", "B", "C"]
        data = [["1", "2", "3"], ["4", "5", "6"]]
        csv_object_original = [dict(zip(header, row)) for row in data]
        csv_path = "test.csv"
        # Write csv file
        file_utils.write_csv_file(csv_object_original, csv_path)

        # Read in csv file
        csv_object = file_utils.read_csv_file(csv_path)

        # Check if the two objects are the same
        self.assertEqual(csv_object_original, csv_object)

        # Delete csv file
        os.remove(csv_path)

    def test_csv_file_functions_with_process_row(self):
        def process_row(row):
            """Convert string to int"""
            return [int(x) for x in row]

        header = ["A", "B", "C"]
        data = [[1, 2, 3], [4, 5, 6]]
        csv_object_original = [dict(zip(header, row)) for row in data]
        csv_path = "test.csv"
        # Write csv file
        file_utils.write_csv_file(csv_object_original, csv_path)

        # Read in csv file
        csv_object = file_utils.read_csv_file(csv_path, process_row_func=process_row)

        # Check if the two objects are the same
        self.assertEqual(csv_object_original, csv_object)

        # Delete csv file
        os.remove(csv_path)


if __name__ == "__main__":
    unittest.main()



import psycopg
import traceback

from typing import Any, List, Dict


class DBConnector:
    # Static attributes
    connector_cache = {}

    def __init__(self, db_id: str):
        self.db_id = db_id
        self.conn = None
        self.cur = None

    def __new__(cls, *args, **kwargs):
        db_id = kwargs.get("db_id", None)
        if not db_id in DBConnector.connector_cache.keys():
            DBConnector.connector_cache[db_id] = super(DBConnector, cls).__new__(cls)
        return DBConnector.connector_cache[db_id]

    def execute(self, sql: str) -> List[Any]:
        return self.cur.execute(sql)

    def fetchall(self) -> List[Any]:
        return self.cur.fetchall()

    def fetchall_with_col_names(self) -> List[Dict[str, Any]]:
        """Return a list of dictionaries with column names as keys and list of values as values"""
        rows = self.fetchall()
        col_names = [col.name for col in self.cur.description]
        assert len(col_names) == len(
            set(col_names)
        ), f"Column names are not unique, len({col_names}) vs len({set(col_names)})"
        assert len(rows) == 0 or len(rows[0]) == len(
            col_names
        ), f"Number of columns ({len(col_names)}) does not match number of rows ({len(rows[0])})"
        # Create a dictionary with column names as keys and list of values as values
        results: List[Dict[str, Any]] = []
        for row in rows:
            # Insert each column value of a row into the corresponding list
            results.append({col_name: value for col_name, value in zip(col_names, row)})
        return results

    def fetchone(self) -> List[Any]:
        return self.cur.fetchone()

    def execute_and_fetchall(self, sql: str) -> List[Any]:
        self.execute(sql)
        return self.fetchall()

    def execute_and_fetchall_with_col_names(self, sql: str) -> List[Dict[str, Any]]:
        self.execute(sql)
        return self.fetchall_with_col_names()

    def close(self) -> None:
        self.conn.close()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self.conn

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
        self.conn.close()
        return True


class PostgresConnector(DBConnector):
    def __init__(self, user_id: str, passwd: str, host: str, port: str, db_id: str):
        super(PostgresConnector, self).__init__(db_id=db_id)
        self.connect(user_id, passwd, host, port, db_id)
        self.conn.autocommit = True

    def __new__(cls, *args, **kwargs):
        return super(PostgresConnector, cls).__new__(cls, *args, **kwargs)

    def connect(self, user_id, passwd, host, port, db_id):
        self.conn = psycopg.connect(
            f"user={user_id} password={passwd} host={host} port={port} dbname={db_id}"
        )
        self.cur = self.conn.cursor()

    def fetch_table_names(self) -> List[str]:
        sql = """
            SELECT *
            FROM pg_catalog.pg_tables
            WHERE schemaname != 'pg_catalog' AND 
                  schemaname != 'information_schema';
        """
        self.cur.execute(sql)
        tables = [
            f"{str(table[0].lower())}.{str(table[1].lower())}".replace("public.", "")
            for table in self.cur.fetchall()
        ]
        return tables

    def fetch_column_names(self, table_ref: str) -> List[str]:
        def table_name_contains_schema(table_ref):
            return "." in table_ref

        if table_name_contains_schema(table_ref):
            table_schema = table_ref.split(".")[0]
            table_name = table_ref.split(".")[1]
        else:
            table_schema = "public"
            table_name = table_ref
        sql = f"""
            SELECT *
            FROM
                information_schema.columns
            WHERE
                table_schema = '{table_schema}'
                AND table_name = '{table_name}';
            """
        self.execute(sql)
        return [str(col[3].lower()) for col in self.fetchall()]

    def fetch_column_types(self, table_ref: str) -> List[str]:
        def table_name_contains_schema(table_ref):
            return "." in table_ref

        if table_name_contains_schema(table_ref):
            table_schema = table_ref.split(".")[0]
            table_name = table_ref.split(".")[1]
        else:
            table_schema = "public"
            table_name = table_ref
        sql = f"""
            SELECT *
            FROM
                information_schema.columns
            WHERE
                table_schema = '{table_schema}'
                AND table_name = '{table_name}';
            """
        self.execute(sql)
        return [str(col[7].lower()) for col in self.fetchall()]



# Import dataclasses according to Python version
import sys

sys_version = sys.version_info
if sys_version[0] == 3 and sys_version[1] >= 7:
    import dataclasses
else:
    try:
        import dataclasses
    except:
        import subprocess

        subprocess.check_call([sys.executable, "-m", "pip", "install", "dataclasses"])
    finally:
        import dataclasses

import functools
from typing import *

import dacite


def asdict(obj: any, skip_none: bool = False) -> Dict:
    def custom_dict_factory(items):
        dict = {}
        for key, value in items:
            if dataclasses.is_dataclass(value):
                dict[key] = asdict(value, skip_none=skip_none)
            else:
                if skip_none and value is None:
                    continue
                dict[key] = value
        return dict

    return dataclasses.asdict(obj, dict_factory=custom_dict_factory)


def from_dict(data_class: dataclasses.dataclass, data: Dict) -> dataclasses.dataclass:
    return dacite.from_dict(data_class, data)


def dataclass(cls=None, **kwargs):
    def wrap(cls):
        cls.dic = property(asdict)
        cls.dic_wo_none = property(lambda self: asdict(self, skip_none=True))
        cls.from_dict = functools.partial(from_dict, data_class=cls)
        return dataclasses.dataclass(cls, **kwargs)

    if cls is None:
        return wrap

    return wrap(cls)


def field(*args, **kwargs):
    """Syntax sugar for dataclasses.field"""
    return dataclasses.field(*args, **kwargs)



import inspect
import json
import logging
import sys
import time as time_module
from contextlib import contextmanager
from typing import Callable, Dict, Generator, List, Optional, Tuple, Union

from hkkang_utils.pattern import SingletonABCMetaWithArgs

# Import dataclasses according to Python version
sys_version = sys.version_info
if sys_version[0] == 3 and sys_version[1] >= 7:
    import dataclasses
else:
    try:
        import dataclasses
    except:
        import subprocess

        subprocess.check_call([sys.executable, "-m", "pip", "install", "dataclasses"])
    finally:
        import dataclasses


# Dataclass
@dataclasses.dataclass
class Period:
    start_time: float
    end_time: float

    @property
    def elapsed_time(self) -> float:
        return self.end_time - self.start_time


# Utility functions
def prettify_time(time_in_sec: float) -> str:
    if time_in_sec < 60:
        return f"{time_in_sec:.4f} seconds"
    return f"{minute(time_in_sec):.0f}min {second(time_in_sec):.4f}sec"


def minute(time_in_sec: float) -> float:
    return time_in_sec / 60


def second(time_in_sec: float) -> float:
    return time_in_sec % 60


# Timer class
class TimerMeta(SingletonABCMetaWithArgs):
    _call_cnt: Dict[type, Dict[str, int]] = dict()

    def __call__(cls, class_name=None, func_name=None):
        # Figure out the class name and function name of the caller
        if class_name is None and func_name is None:
            stack = inspect.stack()
            caller_stack_idx = (
                list(map(lambda k: k.function, stack)).index("__call__") + 1
            )
            func_name = stack[caller_stack_idx][3]
            for idx in range(caller_stack_idx, len(stack)):
                if "self" in stack[idx][0].f_locals:
                    class_name = stack[idx][0].f_locals["self"].__class__.__name__
                    if stack[idx][3] != func_name:
                        func_name = f"{stack[idx][3]}.{func_name}"
                    break
        # Initialize the call count if it is not initialized
        if cls not in cls._call_cnt:
            cls._call_cnt[cls] = dict()
        instance_key = cls.__repr_args__(class_name, func_name)
        if instance_key not in cls._call_cnt[cls]:
            cls._call_cnt[cls][instance_key] = 0
        return super().__call__(class_name, func_name)


class Timer(metaclass=TimerMeta):
    """Timer class to measure the elapsed time of a function.

    Example1 (measure the time of code block):
        # Initialize timer
        timer = Timer()
        timer.start()

        # Perform some task
        ...

        timer.stop()

        # Print the elapsed time
        print(f"Elapsed time: {timer.elapsed_time}")

    Example2 (measure the time of code block within a function):
        def some_function():
            # Code block without measuring time
            ...

            # Initialize timer
            timer = Timer()
            timer.start()

            # Perform some task
            ...

            timer.stop()

            # Code block without measuring time
            ...

        # Call the function
        some_function()
        # Call the function
        some_function()

        # Print the total elapsed time
        print(f"Total Elapsed time: {timer.total_elapsed_time}")
        # Print the average elapsed time
        print(f"Average elapsed time: {timer.average_elapsed_time}")
    """

    def __init__(self, class_name: str, func_name: str):
        # Name is used to identify the timer
        self.class_name = class_name
        self.func_name = func_name
        # Time related variables
        self.start_time: Optional[float] = None
        self.measured_times: List[Period] = []
        self.paused_times: List[Period] = []

    @property
    def name(self) -> str:
        if self.class_name is None:
            return self.func_name
        return f"{self.class_name}.{self.func_name}"

    @property
    def logger(self) -> logging.Logger:
        return logging.getLogger(f"Timer.{self.name}")

    @property
    def call_cnt(self) -> int:
        instance_key = Timer.__repr_args__(self.class_name, self.func_name)
        return Timer._call_cnt[Timer][instance_key]

    @call_cnt.setter
    def call_cnt(self, value):
        instance_key = Timer.__repr_args__(self.class_name, self.func_name)
        Timer._call_cnt[Timer][instance_key] = value

    @property
    def elapsed_time(self) -> float:
        last_period = self.measured_times[-1]
        return last_period.elapsed_time

    @property
    def total_elapsed_time(self) -> float:
        return sum([period.elapsed_time for period in self.measured_times])

    @property
    def avg_elapsed_time(self) -> float:
        if self.call_cnt == 0:
            return 0
        return self.total_elapsed_time / self.call_cnt

    # Methods for measuring the time
    @contextmanager
    def measure(
        self, print_measured_time: bool = False
    ) -> Generator["Timer", None, None]:
        # Pre processing
        self.start()
        yield self
        self.stop()
        # Post processing
        if print_measured_time:
            self.show_elapsed_time()

    @contextmanager
    def pause(self) -> Generator["Timer", None, None]:
        # Pre processing
        current_time = time_module.time()
        self.measured_times.append(Period(self.start_time, current_time))
        self.start_time = current_time
        yield self
        # Post processing
        current_time = time_module.time()
        self.paused_times.append(Period(self.start_time, current_time))
        self.start_time = current_time

    def start(self) -> None:
        if self.start_time is not None:
            self.logger.warning(
                "Timer has already been started. Please call stop() first."
            )
        # Increase the call count
        self.call_cnt += 1
        self.start_time = time_module.time()

    def stop(self) -> float:
        current_time = time_module.time()
        if self.start_time is None:
            self.logger.warning(
                "Timer has not been started. Please call start() first."
            )
        self.measured_times.append(Period(self.start_time, current_time))
        self.start_time = None
        return self.elapsed_time

    # Methods for printing the measured time
    def show_elapsed_time(self) -> None:
        self.logger.info(f"Elapsed time: {prettify_time(self.elapsed_time)}")

    def show_total_elapsed_time(self) -> None:
        self.logger.info(
            f"Total elapsed time: {prettify_time(self.total_elapsed_time)}"
        )

    def show_avg_elapsed_time(self) -> None:
        self.logger.info(
            f"Average elapsed time: {prettify_time(self.avg_elapsed_time)}"
        )

    def summarize_measured_time(
        self, silent: bool = False, in_dict: bool = False
    ) -> Union[Dict[str, Union[str, int, float]], Tuple[str, int, float, float]]:
        if in_dict:
            result = {
                "name": self.name,
                "call_cnt": self.call_cnt,
                "avg_elapsed_time": self.avg_elapsed_time,
                "total_elapsed_time": self.total_elapsed_time,
            }
            if not silent:
                self.logger.info(json.dumps(result, indent=4))
        else:
            result = (
                self.name,
                self.call_cnt,
                self.avg_elapsed_time,
                self.total_elapsed_time,
            )
            if not silent:
                self.logger.info(f"Function name: {self.name}")
                self.logger.info(f"Call count: {self.call_cnt}")
                self.show_avg_elapsed_time()
                self.show_total_elapsed_time()
        return result

    @classmethod
    def summarize_measured_times(
        cls, silent: bool = False, in_dict: bool = False
    ) -> List[Tuple[str, int, float, float]]:
        return_values = []
        for timer in Timer._instances[cls].values():
            return_values.append(
                timer.summarize_measured_time(silent=silent, in_dict=in_dict)
            )
        return return_values


# Decorator for timer
def measure_time(func: Callable, print_measured_time: bool = True) -> Callable:
    """Decorator to measure the elapsed time of a function.

    :param func: function to be measured
    :type func: Callable
    :param print_measured_time: whether to print the measured time after the function call, defaults to True
    :type print_measured_time: bool, optional

    Example:
        @measure_time
        def some_function():
            # All the code block within this function will be measured
            ...

        # Calling the function will print the measured time
        some_function()
    """

    def wrapper(*args, **kwargs):
        names = list(filter(lambda n: n != "<locals>", func.__qualname__.split(".")))
        class_name = None if len(names) == 1 else names[0]
        func_name = names[0] if len(names) == 1 else ".".join(names[1:])
        timer = Timer(class_name=class_name, func_name=func_name)
        with timer.measure(print_measured_time=print_measured_time):
            result = func(*args, **kwargs)
        return result

    return wrapper


if __name__ == "__main__":
    timer = Timer()
    print("Done")



import functools
import math
import operator
from typing import *

from tqdm import tqdm


def do_flatten_list(list_of_list: List[List[Any]]) -> List[Any]:
    """Flatten a list of list to a list

    :param list_of_list: list to flatten
    :type list_of_list: List[List[Any]]
    :return: flatten list
    :rtype: List[Any]
    """
    return functools.reduce(operator.iconcat, list_of_list, [])


def map_many(functions: List, iterable: List[Any]) -> List[Any]:
    return list(functools.reduce(lambda x, y: map(y, x), functions, iterable))


def get(items: List[Any], idx: int, default: Any = "") -> Any:
    if idx < 0:
        return default
    try:
        return items[idx]
    except IndexError:
        return default


def divide_into_chunks(
    lst: Union[List[Any], Dict], num_chunks: int
) -> Union[List[Any], Dict]:
    num_of_items_in_chunk = math.ceil(len(lst) / num_chunks)
    if isinstance(lst, dict):
        # Convert dict items to a list of key-value tuples, split them, and then convert back to a list of dictionaries
        items = list(lst.items())
        divided_chunks = [
            dict(items[i : i + num_of_items_in_chunk])
            for i in range(0, len(items), num_of_items_in_chunk)
        ]
    elif isinstance(lst, list):
        divided_chunks = [
            lst[i : i + num_of_items_in_chunk]
            for i in range(0, len(lst), num_of_items_in_chunk)
        ]
    return divided_chunks


def chunks(
    iterator: Union[Iterable[Any], List[Any]],
    chunk_size: int,
    show_progress: bool = False,
) -> Iterable[List[Any]]:
    """Yield successive n-sized chunks from iterator."""
    if show_progress:
        iterator = tqdm(iterator)
    chunk = []
    for x in iterator:
        chunk.append(x)
        if len(chunk) >= chunk_size:
            yield chunk
            chunk = []
    if len(chunk) > 0:
        yield chunk


def safe_zip(*args: List[Iterable]) -> Iterable[tuple]:
    item_len = 0
    for idx, arg in enumerate(args):
        if idx == 0:
            # Save the length of first item
            item_len = len(arg)
        else:
            # Check if all items have the same length
            if len(arg) != item_len:
                raise ValueError("All arguments must have the same length")
    return zip(*args)



import functools
import logging
import pprint
import re
from typing import *

import pglast
from sqlfluff.core import Linter, SQLBaseError, SQLLexError, SQLParseError
from sqlfluff.core.parser.segments.base import UnparsableSegment

from hkkang_utils.string import multi_space_to_single_space


def double_quotes_to_single_quotes(sql_str):
    return sql_str.replace('"', "'")


def add_subquery_alias(sql_str):
    tokens_before_parenthesis = []
    last_token = ""
    current_token = ""
    alias_cnt = 1
    parenthesis_depth = 0
    sql_str = multi_space_to_single_space(sql_str)
    new_sql_str = ""
    for character in sql_str:
        new_sql_str += character
        # State change
        if character in [" ", "\n", "\t"]:
            last_token = current_token
            current_token = ""
        else:
            current_token += character
        # Set parenthesis depth
        if character == "(":
            parenthesis_depth += 1
            tokens_before_parenthesis.append(last_token)
        elif character == ")":
            parenthesis_depth -= 1
            popped_token = tokens_before_parenthesis.pop(-1)
            if popped_token.lower() == "from":
                # Add alias
                new_alias = f"S{alias_cnt}"
                alias_cnt += 1
                new_sql_str += f" AS {new_alias}"
    return new_sql_str


def replace_table_or_column_names_with_postgresql_keywords(sql_str):
    keywords = ["CAST", "cast"]
    for keyword in keywords:
        sql_str = sql_str.replace(f" FROM {keyword} ", f" FROM {keyword}1 ").replace(
            f" JOIN {keyword} ", f" JOIN {keyword}1 "
        )
    return sql_str


def loosely_parse(sql_str):
    """_summary_: Parse a SQL string and return a parse tree. If the parsing fails, return None.

    Args:
        sql_str (str): SQL string to parse

    Returns:
        pglast.ast.Node: pglast Node object
    """

    def is_nested_query_in_from_clause_without_alias(e_args):
        return e_args[0] == "subquery in FROM must have an alias"

    def has_join_keyword_without_join_condition(sql, e_args):
        sql = sql.lower()
        if "syntax error at" in e_args[0].lower():
            # Cut off the WHERE clause
            where_start_idx = e_args[1]
            sql_substring = sql[:where_start_idx]
            # Cut off before FROM clause
            from_start_idx = sql_substring.index("from")
            sql_substring = sql_substring[from_start_idx:]
            if " join " in sql_substring and " on " not in sql_substring:
                return True
        return False

    def add_natural_keyword(sql, where_start_idx):
        sql_front = sql[:where_start_idx]
        sql_rear = sql[where_start_idx:]
        return (
            sql_front.replace(" join ", " natural join ").replace(
                " JOIN ", " NATURAL JOIN "
            )
            + sql_rear
        )

    sql_str_ = double_quotes_to_single_quotes(sql_str)
    sql_str_ = replace_table_or_column_names_with_postgresql_keywords(sql_str_)
    try:
        parse_tree = pglast.parse_sql(sql_str_)
    except pglast.parser.ParseError as e:
        if is_nested_query_in_from_clause_without_alias(e.args):
            sql_str_ = add_subquery_alias(sql_str_)
            return loosely_parse(sql_str_)
        elif has_join_keyword_without_join_condition(sql_str_, e.args):
            sql_str_ = add_natural_keyword(sql_str_, where_start_idx=e.args[1])
            return loosely_parse(sql_str_)
        else:
            raise e
    return parse_tree[0]


# Recursively set att_name of object to att_value
def set_att_value(object, att_name, att_value):
    def is_pglast_object(obj):
        return str(type(obj)).startswith("<class 'pglast.")

    if hasattr(object, att_name):
        setattr(object, att_name, att_value)
    for att_str in filter(
        lambda att_str: not att_str.startswith("__")
        and (
            type(getattr(object, att_str)) in [list, tuple]
            or is_pglast_object(getattr(object, att_str))
        ),
        dir(object),
    ):
        selected_object = getattr(object, att_str)
        if type(selected_object) in [list, tuple]:
            for item in selected_object:
                set_att_value(item, att_name, att_value)
        else:
            set_att_value(selected_object, att_name, att_value)
    return object


def pprint_parse_tree(parse_tree):
    """_summary_: Print a parse tree in a readable json format.

    Args:
        parse_tree (pglast.ast.Node): pglast Node object
    """

    def handle_null(input_str):
        return input_str.replace("None", "null")

    def handle_boolean(input_str):
        return input_str.replace(" True", " true").replace(" False", " false")

    def handle_tuple(input_str):
        return input_str.replace("(", "[").replace(",)", "]").replace(")", "]")

    def handle_qutoes(input_str):
        return input_str.replace("'", '"')

    string_to_print = pprint.pformat(parse_tree)
    processed_string = functools.reduce(
        lambda x, func: func(x),
        [handle_null, handle_boolean, handle_tuple, handle_qutoes],
        string_to_print,
    )
    print(processed_string)


### Prettify SQL

# Regex for CREATE TABLE statements
CREATE_TABLE_RE = re.compile(r"(?is)^(CREATE TABLE\s+\S+\s*)\((.*)\)(;?)$")
# Regex for UPSERT (PostgreSQL) statements
UPSERT_RE = re.compile(
    r"(?is)^(INSERT\s+INTO\s+\S+\s*)\((.*?)\)\s*VALUES\s*\((.*?)\)\s*(ON\s+CONFLICT.*?)(;?)$",
    re.DOTALL,
)


def _split_top_level(text: str) -> List[str]:
    """Split a comma-separated list at top-level parentheses, ignoring commas inside quotes."""
    items: List[str] = []
    buffer: List[str] = []
    depth = 0
    in_sq = False
    in_dq = False
    i = 0
    while i < len(text):
        ch = text[i]
        if ch == "'" and not in_dq:
            in_sq = not in_sq
            buffer.append(ch)
        elif ch == '"' and not in_sq:
            in_dq = not in_dq
            buffer.append(ch)
        elif not in_sq and not in_dq:
            if ch == "(":
                depth += 1
                buffer.append(ch)
            elif ch == ")":
                depth -= 1
                buffer.append(ch)
            elif ch == "," and depth == 0:
                items.append("".join(buffer).strip())
                buffer = []
            else:
                buffer.append(ch)
        else:
            buffer.append(ch)
        i += 1
    if buffer:
        items.append("".join(buffer).strip())
    return items


def _format_block(header: str, body: str, closing: str = ")") -> str:
    """Format a SQL block with parentheses and comma-separated elements."""
    body = re.sub(r"\s+", " ", body).strip()
    fields = _split_top_level(body)
    last = len(fields) - 1
    lines = [f"    {field}{',' if i < last else ''}" for i, field in enumerate(fields)]
    return f"{header}(\n" + "\n".join(lines) + f"\n){closing}"


def _format_assignments(conflict_clause: str) -> str:
    """
    Split ON CONFLICT ... DO UPDATE SET assignments and format them.
    Returns a formatted string including the assignments.
    """
    m = re.match(
        r"(?is)^(ON\s+CONFLICT\s*\(.*?\)\s*DO\s+UPDATE\s+SET)\s*(.*)$", conflict_clause
    )
    if not m:
        return conflict_clause
    header, assigns = m.groups()
    assigns = assigns.rstrip()
    parts = _split_top_level(assigns)
    last = len(parts) - 1
    lines = [f"    {part}{',' if i < last else ''}" for i, part in enumerate(parts)]
    return f"{header}\n" + "\n".join(lines)


def prettify_sql(sql: str) -> str:
    """Prettify CREATE TABLE and UPSERT (INSERT ... ON CONFLICT) SQL statements."""
    sql = sql.strip()
    # Normalize whitespace for matching
    compact = re.sub(r"\s+", " ", sql)
    # CREATE TABLE
    m = CREATE_TABLE_RE.match(compact)
    if m:
        header, body, closing = m.groups()
        return _format_block(header, body, closing)
    # UPSERT
    m = UPSERT_RE.match(sql)
    if m:
        insert_header, cols, vals, conflict_clause, closing = m.groups()
        cols_block = _format_block(insert_header, cols, closing="")
        vals_block = _format_block("VALUES", vals, closing="")
        conflict_block = _format_assignments(
            re.sub(r"\s+", " ", conflict_clause).strip()
        )
        return "\n".join([cols_block, vals_block, conflict_block + closing])
    # Fallback: return unchanged
    return sql


### Validate SQL string
def is_valid_sql(sql_string: str, dialect: str = "postgres") -> bool:
    """
    Return True if sqlfluff can parse the given SQL string without any
    'unparsable' segments (i.e., it’s syntactically valid for the chosen dialect).
    Return False if parsing fails or if any UnparsableSegment is present.
    """
    sqlfluff_logger = logging.getLogger("sqlfluff")
    prev_level = sqlfluff_logger.level
    sqlfluff_logger.setLevel(logging.WARNING)

    linter = Linter(dialect=dialect)
    try:
        parsed = linter.parse_string(sql_string)
    except (SQLParseError, SQLLexError, SQLBaseError) as e:
        logging.warning(f"sqlfluff raised a parsing error: {e}")
        return False
    finally:
        sqlfluff_logger.setLevel(prev_level)

    # If parse_string succeeds but yields no parse tree at all, it’s invalid
    if not getattr(parsed, "tree", None):
        return False

    # Walk the entire tree and reject if any UnparsableSegment appears
    for segment in parsed.tree.iter_segments():
        if isinstance(segment, UnparsableSegment):
            logging.warning(f"Found unparsable segment in SQL: {sql_string!r}")
            return False

    return True



import os
import random
from typing import List, Optional, Tuple

import numpy as np
import torch


def show_environment_setting(print_fn):
    msg = "CUDA INFO:"
    msg += f"\n\tIs CUDA supported by this system? {torch.cuda.is_available()}"
    if torch.cuda.is_available():
        msg += f"\n\tCUDA version: {torch.version.cuda}"
        msg += f"\n\tID of current CUDA device: {torch.cuda.current_device()}"
        msg += f"\n\tName of current CUDA device: {torch.cuda.get_device_name(torch.cuda.current_device())}"
    print_fn(msg)


def att_mask_from_individual_masks(
    tgt_mask: torch.Tensor, src_mask: torch.Tensor
) -> torch.Tensor:
    """Combine two individual masks for attention mask

    :param tgt_mask: mask for target sequence with shape (bsz, tgt_len)
    :type tgt_mask: torch.Tensor
    :param src_mask: mask for source sequence with shape (bsz, src_len)
    :type src_mask: torch.Tensor
    :return: attention mask with shape (bsz, tgt_len, src_len)
    :rtype: torch.Tensor
    """
    tgt_len = tgt_mask.shape[-1]
    src_len = src_mask.shape[-1]

    _tgt_mask = tgt_mask.unsqueeze(-1).repeat(1, 1, src_len)
    _src_mask = src_mask.unsqueeze(1).repeat(1, tgt_len, 1)

    return _tgt_mask * _src_mask


def zero_pad_batching_one_dim(
    tensor_list: List[torch.Tensor], return_mask=True
) -> Tuple[torch.Tensor, torch.Tensor]:
    max_len = max([len(tensor) for tensor in tensor_list])
    device = tensor_list[0].device
    shape = (
        (len(tensor_list), max_len)
        if len(tensor_list[0].shape) == 1
        else (len(tensor_list), max_len, *list(tensor_list[0].shape[1:]))
    )
    token_tensor = torch.zeros(shape, dtype=tensor_list[0].dtype, device=device)
    if return_mask:
        mask_tensor = torch.ones(shape, dtype=torch.bool, device=device)
    for idx, tensor in enumerate(tensor_list):
        token_tensor[idx, : len(tensor)] = tensor
        if return_mask:
            mask_tensor[idx, : len(tensor)] = False
    return (token_tensor, mask_tensor) if return_mask else token_tensor


def zero_pad_batching_two_dim(
    tensor_list: List[torch.Tensor], return_mask=True
) -> torch.Tensor:
    device = tensor_list[0].device
    dim_one_max_len = max([len(tensor) for tensor in tensor_list])
    dim_two_max_len = max([len(tensor[0]) for tensor in tensor_list])
    token_tensor = torch.zeros(
        len(tensor_list),
        dim_one_max_len,
        dim_two_max_len,
        dtype=tensor_list[0].dtype,
        device=device,
    )
    if return_mask:
        mask_tensor = torch.ones(token_tensor, dtype=torch.bool, device=device)
    for idx, tensor in enumerate(tensor_list):
        token_tensor[idx, : len(tensor), : len(tensor[0])] = tensor
        if return_mask:
            mask_tensor[idx, : len(tensor), : len(tensor[0])] = False
    return (token_tensor, mask_tensor) if return_mask else token_tensor


def zero_pad_batching(tensor_list: List[torch.Tensor]) -> torch.Tensor:
    tensor_item = tensor_list[0]
    if len(tensor_item.shape) == 1:
        return zero_pad_batching_one_dim(tensor_list)
    elif len(tensor_item.shape) == 2:
        return zero_pad_batching_two_dim(tensor_list)
    else:
        raise NotImplementedError(
            f"Only support 1D and 2D tensor, but found: {tensor_item.shape}"
        )


def set_torch_deterministic(seed: Optional[int] = 0) -> None:
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # if you are using multi-GPU.
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True


def is_torch_compile_available() -> None:
    if torch.cuda.is_available():
        device_cap = torch.cuda.get_device_capability()
        return device_cap[0] >= 7
    return False


def calculate_model_size(model: torch.nn.Module, in_MB: bool = False) -> float:
    """Compute model size in KB or MB

    :param model: model to compute size
        in_MB: if True, return size in MB, otherwise return size in KB
    :type model: torch.nn.Module
    :return: model size in KB or MB
    :rtype: int
    """
    param_size = 0
    for param in model.parameters():
        param_size += param.nelement() * param.element_size()
    buffer_size = 0
    for buffer in model.buffers():
        buffer_size += buffer.nelement() * buffer.element_size()

    # Default is in KB
    exp_val = 2 if in_MB else 3
    return (param_size + buffer_size) / 1024**exp_val


def count_parameters(model: torch.nn.Module, only_trainable: bool = True) -> int:
    """Count the number of trainable model parameter

    :param model: model to compute size
    :type model: torch.nn.Module
    :return: number of model parameter
    :rtype: int
    """
    return sum(
        p.numel() for p in model.parameters() if p.requires_grad or (not only_trainable)
    )


def print_trainable_parameters(model):
    """
    Prints the number of trainable parameters in the model.
    """
    trainable_params = 0
    all_param = 0
    for _, param in model.named_parameters():
        all_param += param.numel()
        if param.requires_grad:
            trainable_params += param.numel()
    print(
        f"trainable params: {trainable_params} || all params: {all_param} || trainable%: {100 * trainable_params / all_param}"
    )


if __name__ == "__main__":
    show_environment_setting(print)



import sys

# Import dataclasses according to Python version
sys_version = sys.version_info
if sys_version[0] == 3 and sys_version[1] >= 7:
    import dataclasses
else:
    try:
        import dataclasses
    except:
        import subprocess

        subprocess.check_call([sys.executable, "-m", "pip", "install", "dataclasses"])
    finally:
        import dataclasses

import unittest
from typing import *

import src.hkkang_utils.data as data_utils


@dataclasses.dataclass
class TmpDataclass:
    a: int
    b: str
    c: Optional[int] = dataclasses.field(default=None)
    d: Optional[str] = dataclasses.field(default=None)
    e: List[int] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class TmpNestedDataclass:
    name: str
    nested: TmpDataclass


class Test_data_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_data_utils, self).__init__(*args, **kwargs)
        self.data_class = TmpDataclass(1, "a", None, None, [1, 2, 3])
        self.nested_data_class = TmpNestedDataclass("tmp_name", self.data_class)
        # Gold dicts
        self.data_class_dic_wo_none = {"a": 1, "b": "a", "e": [1, 2, 3]}
        self.data_class_dic_w_none = {
            "a": 1,
            "b": "a",
            "c": None,
            "d": None,
            "e": [1, 2, 3],
        }
        self.nested_data_class_dic_wo_none = {
            "name": "tmp_name",
            "nested": self.data_class_dic_wo_none,
        }
        self.nested_data_class_dic_w_none = {
            "name": "tmp_name",
            "nested": self.data_class_dic_w_none,
        }

    def test_asdict(self):
        generated_nested_data_class_dict_w_none = data_utils.asdict(
            self.nested_data_class, skip_none=False
        )
        generated_nested_data_class_dict_wo_none = data_utils.asdict(
            self.nested_data_class, skip_none=True
        )

        self.assertEqual(
            generated_nested_data_class_dict_w_none,
            self.nested_data_class_dic_w_none,
            f"Generated:{generated_nested_data_class_dict_w_none}\nGold:{self.nested_data_class_dic_w_none}",
        )
        self.assertEqual(
            generated_nested_data_class_dict_wo_none,
            self.nested_data_class_dic_wo_none,
            f"Generated:{generated_nested_data_class_dict_wo_none}\nGold:{self.nested_data_class_dic_wo_none}",
        )

    def test_from_dict(self):
        generated_nested_data_class_from_w_none = data_utils.from_dict(
            TmpNestedDataclass, self.nested_data_class_dic_w_none
        )
        generated_nested_data_class_from_wo_none = data_utils.from_dict(
            TmpNestedDataclass, self.nested_data_class_dic_wo_none
        )
        self.assertEqual(
            generated_nested_data_class_from_w_none,
            self.nested_data_class,
            f"Generated:{generated_nested_data_class_from_w_none}\nGold:{self.nested_data_class}",
        )
        self.assertEqual(
            generated_nested_data_class_from_wo_none,
            self.nested_data_class,
            f"Generated:{generated_nested_data_class_from_wo_none}\nGold:{self.nested_data_class}",
        )


if __name__ == "__main__":
    unittest.main()



import contextlib
import logging
import os
from typing import *

from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from slack_sdk.web.slack_response import SlackResponse
import hkkang_utils.misc as misc_utils
import hkkang_utils.socket as socket_utils

# Load environment variables
misc_utils.load_dotenv(stack_depth=2)

DEFAULT_ACCESS_TOKEN = os.environ.get("SLACK_ACCESS_TOKEN")

logger = logging.getLogger("SlackMessenger")


class SlackMessenger:
    """A wrapper class for sending Slack messages with enhanced functionality.

    This class provides a convenient interface for sending messages to Slack channels,
    managing message threads, and handling notifications with success/error states.

    Attributes:
        channel (str): The Slack channel to send messages to
        token (str): Slack API authentication token
        append_server_info (bool): Whether to include server information in messages
        last_thread_ts (str): Timestamp of the last message thread
    """

    def __init__(
        self,
        channel: str,
        token: str = DEFAULT_ACCESS_TOKEN,
        web_client: Optional[WebClient] = None,
        append_server_info: bool = True,
    ):
        """Initialize the SlackMessenger.

        Args:
            channel: The Slack channel to send messages to
            token: Slack API authentication token (defaults to env variable)
            web_client: Optional pre-configured WebClient instance
            append_server_info: Whether to include server information in messages
        """
        self.channel = channel
        self.token = token
        self._web_client = web_client
        self.append_server_info = append_server_info
        self.last_thread_ts = None
        self.__post_init__()

    def __post_init__(self):
        if not self.token:
            raise ValueError(
                """Please set token or SLACK_ACCESS_TOKEN environment variable.
                   Follow the tutorial to set up bot OAuthToken and permissions: 
                   https://github.com/slackapi/python-slack-sdk/tree/main/tutorial"""
            )

    @property
    def client(self) -> WebClient:
        """Get or create a WebClient instance.

        Returns:
            WebClient: The Slack Web API client
        """
        if self._web_client is None:
            self._web_client = WebClient(token=self.token)
        return self._web_client

    @contextlib.contextmanager
    def notification(
        self,
        success_msg: str,
        error_msg: str,
        start_msg: Optional[str] = None,
        user_id_to_mention: Optional[str] = None,
        replies: Optional[List[str]] = None,
        disable: bool = False,
        disable_callback: Optional[Callable[[], bool]] = None,
    ) -> Generator[Optional["SlackMessenger"], None, None]:
        """
        Send a message when the task within the code block is finished, with an optional short preview.
        If start_msg is None:
            1. The success/error message will be sent.
            2. The replies will be sent as replies to the success/error message.
        If start_msg is not None:
            1. The start message will be sent.
            2. The replies will be sent as replies to the start message.
            3. The success/error message will be sent as a reply to the start message.
        """
        # Skip notification if the debugger is active or if the disable flag is True
        if misc_utils.is_debugger_active() or disable:
            yield None
            return

        # Append the user_id_to_mention to the messages if it is provided
        if user_id_to_mention is not None:
            success_msg = f"<@{user_id_to_mention}> {success_msg}"
            error_msg = f"<@{user_id_to_mention}> {error_msg}"
            if start_msg is not None:
                start_msg = f"<@{user_id_to_mention}> {start_msg}"

        thread_ts = None  # Initialize thread_ts here
        try:
            # Skip notification if the disable_callback returns True
            if disable_callback is not None and disable_callback():
                yield None
                return

            # Send the start message if it is not None
            if start_msg is not None:
                response = self.send(
                    start_msg, replies=replies
                )  # Use self.send instead of send_message directly
                thread_ts = response["ts"]
                self.last_thread_ts = thread_ts

            yield self

            # Send the success message
            if start_msg is None:
                self.send(success_msg, replies=replies)
            else:
                # Send the start message as reply to the start message
                self.send_reply(success_msg, thread_ts=thread_ts)
        except Exception as e:
            if disable_callback is not None and disable_callback():
                return
            # Send the error message
            message_to_send = f"{error_msg} ({e.__class__.__name__}: {e})"
            if start_msg is None:
                self.send(message_to_send, replies=replies)
            else:
                # Send the error message as reply to the start message
                self.send_reply(message_to_send, thread_ts=thread_ts)
            raise

    def send(self, text: str, replies: Optional[List[str]] = None) -> SlackResponse:
        """Send a Slack message with optional replies."""
        response = send_message(
            channel=self.channel,
            text=text,
            web_client=self.client,
            token=self.token,
            append_server_info=self.append_server_info,
        )
        self.last_thread_ts = response["ts"]

        # Send replies if provided
        if replies:
            for reply in replies:
                self.send_reply(reply, thread_ts=self.last_thread_ts)

        return response

    def send_reply(self, text: str, thread_ts: Optional[str] = None) -> SlackResponse:
        """Send a Slack reply with an optional reply."""
        if thread_ts is None:
            assert (
                self.last_thread_ts is not None
            ), "No thread to reply to. Send a message first."
            thread_ts = self.last_thread_ts

        response = send_reply(
            channel=self.channel,
            text=text,
            web_client=self.client,
            token=self.token,
            thread_ts=thread_ts,
        )
        self.last_thread_ts = response["ts"]
        return response


def _get_or_create_client(
    web_client: Optional[WebClient] = None, token: str = DEFAULT_ACCESS_TOKEN
) -> WebClient:
    """Create a new WebClient or return the provided one.

    Args:
        web_client: Existing WebClient instance, if any
        token: Slack API token to use when creating a new client

    Returns:
        WebClient: A configured Slack Web API client

    Raises:
        AssertionError: If no token is provided when creating a new client
    """
    if web_client is None:
        assert token is not None, "Please set SLACK_ACCESS_TOKEN environment variable."
        web_client = WebClient(token=token)
    return web_client


def send_message(
    channel: str,
    text: str,
    web_client: Optional[WebClient] = None,
    token: str = DEFAULT_ACCESS_TOKEN,
    append_server_info: bool = True,
) -> SlackResponse:
    """Send a single message to Slack.

    Args:
        channel: The channel to send the message to
        text: The message content
        web_client: Optional pre-configured WebClient
        token: Slack API token
        append_server_info: Whether to include server information

    Returns:
        SlackResponse: The response from the Slack API

    Raises:
        SlackApiError: If the message cannot be sent
    """
    web_client = _get_or_create_client(web_client, token)

    if append_server_info:
        ip = socket_utils.get_local_ip()
        host_name = socket_utils.get_host_name()
        text_with_prefix = f"Message from {host_name} ({ip}):\n{text}"
    else:
        text_with_prefix = text

    try:
        response = web_client.chat_postMessage(
            channel=channel,
            text=text_with_prefix,
        )
        return response
    except SlackApiError as e:
        error_msg = f"Error sending text message: {e.response['error']}"
        logger.error(error_msg)
        raise SlackApiError(message=error_msg, response=e.response)


def send_reply(
    channel: str,
    thread_ts: str,
    text: str,
    web_client: Optional[WebClient] = None,
    token: str = DEFAULT_ACCESS_TOKEN,
) -> SlackResponse:
    """Send a Slack reply with an optional reply."""
    web_client = _get_or_create_client(web_client, token)

    # Send the reply
    try:
        response = web_client.chat_postMessage(
            channel=channel,
            thread_ts=thread_ts,
            text=text,
        )
        return response
    except SlackApiError as e:
        error_msg = f"Error sending reply: {e.response['error']}"
        logger.error(error_msg)
        raise SlackApiError(message=error_msg, response=e.response)


@contextlib.contextmanager
def notification(
    channel: str,
    success_msg: str,
    error_msg: str,
    start_msg: Optional[str] = None,
    user_id_to_mention: Optional[str] = None,
    token: str = DEFAULT_ACCESS_TOKEN,
    web_client: Optional[WebClient] = None,
    replies: Optional[List[str]] = None,
    disable: bool = False,
    disable_callback: Optional[Callable[[], bool]] = None,
    append_server_info: bool = True,
) -> Generator[Optional["SlackMessenger"], None, None]:
    """Send a message when the task within the code block is finished, with an optional short preview."""
    slack_messenger = SlackMessenger(
        channel=channel,
        token=token,
        web_client=web_client,
        append_server_info=append_server_info,
    )

    with slack_messenger.notification(
        success_msg=success_msg,
        error_msg=error_msg,
        start_msg=start_msg,
        user_id_to_mention=user_id_to_mention,
        replies=replies,
        disable=disable,
        disable_callback=disable_callback,
    ) as messenger:
        yield messenger


@contextlib.contextmanager
def slack_notification(*args, **kwargs):
    raise NotImplementedError("Please use notification instead of slack_notification")



import wandb


def push(
    project: str,
    dataset: str,
    data_type: str,
    data_dir_path: str,
):
    """_summary_

    :param project: Name of the project
    :type project: str
    :param dataset: Name of the dataset
    :type dataset: str
    :param data_type: It should be one of dataset or result
    :type data_type: str
    :param data_dir_path: data directory path
    :type data_dir_path: str
    """
    assert data_type in [
        "dataset",
        "result",
    ], f"data_type should be one of dataset or result, but {data_type}"
    run = wandb.init(project=project, job_type="add-dataset")
    artifact = wandb.Artifact(name=dataset, type=data_type)
    artifact.add_dir(local_path=data_dir_path)
    run.log_artifact(artifact)
    run.finish()


def pull(project: str, dataset: str, alias: str = "latest", data_dir_path: str = None):
    api = wandb.Api()
    artifact = api.artifact(f"{project}/{dataset}:{alias}")
    artifact_dir = artifact.download(data_dir_path)
    return artifact_dir



import functools
import logging
import threading
import traceback
from concurrent.futures import ProcessPoolExecutor
from typing import *

from hkkang_utils.list import divide_into_chunks, do_flatten_list


def _shorten_string(text: str) -> str:
    assert isinstance(text, str), f"text must be str type, but {type(text)} is given."
    if len(text) > 10:
        return text[:10] + "..."
    return text


class Thread(threading.Thread):
    """Please use start method to start thread. start method will call run method.
    If you are doing cpu intensive task, please use MultiProcessor class instead of this class. It's much faster.

    Example:
        # Initialize threads
        thread1 = concurrent_utils.Thread(count_million_and_print_name, 1, {"name":"name_1"})
        thread2 = concurrent_utils.Thread(count_million_and_print_name, 2, {"name":"name_2"})

        # Start threads
        thread1.start()
        thread2.start()

        # Wait for threads to finish (this is optional: accessing result will call join method anyway)
        thread1.join()
        thread2.join()

        # Get results
        result1 = thread1.result
        result2 = thread2.result
    """

    # Static variable
    cnt = 0

    def __init__(self, *args, **kwargs):
        super().__init__()
        Thread.cnt += 1
        self.threadID = Thread.cnt
        self.func = args[0]
        self.args = args[1:]
        self.kwargs = kwargs
        self._result = None
        self.logger = logging.getLogger(f"Thread {self.threadID}")

    @property
    def result(self):
        if self.is_alive():
            self.join()
        return self._result

    def run(self):
        post_fix = f"Thread {self.threadID}: {self.func.__name__}(args={_shorten_string(str(self.args))}, kwargs={_shorten_string(str(self.kwargs))})"
        self.logger.info(f"Starting {post_fix}")
        self._result = self.func(*self.args, **self.kwargs)
        self.logger.info(f"Exiting {post_fix}")

    def _parse_args(self, args):
        if args is None:
            return ()
        elif isinstance(args, tuple):
            return args
        elif isinstance(args, list):
            return tuple(args)
        else:
            return (args,)

    def _parse_kwargs(self, kwargs):
        if kwargs is None:
            return {}
        elif isinstance(kwargs, dict):
            return kwargs
        else:
            raise ValueError(f"kwargs must be dict type, but {type(kwargs)} is given.")


class MultiProcessor:
    """
    example:
        # Initialize multi processor
        multiprocessor = Multiprocessor(num_process=4)

        # Run processes concurrently
        multiprocessor.run(count_million_and_print_name, 1, name=f"name_{i}")
        multiprocessor.run(count_million_and_print_name, 2, name=f"name_{i}")
        multiprocessor.run(count_million_and_print_name, 3, name=f"name_{i}")
        multiprocessor.run(count_million_and_print_name, 4, name=f"name_{i}")

        # Wait for all processes to finish (this is optional: accessing result will call join method anyway)
        multiprocessor.join()

        # Get results
        results = multiprocessor.results
    """

    def __init__(self, num_workers: int, store_results: bool = True):
        self.num_workers = num_workers
        self._futures = []
        self._results = []
        self.store_results = store_results
        self.logger = logging.getLogger(f"MultiProcessor")

    @property
    def results(self):
        return self._results

    @functools.cached_property
    def executor(self):
        return ProcessPoolExecutor(max_workers=self.num_workers)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        # Clean up: Wait for all processes to finish
        self.join()

        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
            return False
        return True

    def _check_if_valid_function(self, func: Callable):
        if not hasattr(func, "__call__"):
            self.logger.error(
                f"Given function is not callable. Please use callable function."
            )
        elif func.__name__ == "<lambda>":
            self.logger.error(
                f"Given function is lambda function. Please use normal function instead of lambda function."
            )
        else:
            return True
        return False

    def run(self, *args, **kwargs):
        """Pass callable function in the first argument.
        Then, pass other arguments and keyword arguments as you would to the callable function.
        """
        # Extract function and arguments
        assert len(args) > 0, f"Please pass function as the first argument."
        func: Callable = args[0]
        args = args[1:]

        # Check the given function is not lambda function
        self._check_if_valid_function(func)

        self._futures.append(
            self.executor.submit(functools.partial(func, *args, **kwargs))
        )

    def join(self):
        if self.store_results:
            self._results = [future.result() for future in self._futures]
        else:
            for future in self._futures:
                future.result()
        # Free memory
        self.executor.shutdown()


class PartialProcessor:
    """
    This is a class to process data partially.
    Use this class when you want to call multiple programs in terminal to process data in parallel.
    It will split the data and process the data partially considering the total number of processes and the current process number.
    """

    def __init__(
        self,
        func: Optional[Callable] = None,
        total_proc_n: Optional[int] = None,
        current_proc_n: Optional[int] = None,
    ):
        self.func = func
        self.total_proc_n = total_proc_n
        self.current_proc_n = current_proc_n

    def __call__(
        self,
        data: Union[List[Any], Dict],
        func: Optional[Callable] = None,
        total_proc_n: Optional[int] = None,
        current_proc_n: Optional[int] = None,
    ) -> Any:
        func = self._get_func(func)
        assert isinstance(
            data, list
        ), f"data must be list type, but {type(data)} is given."

        # Divide the data into chunks (each chunk will be processed by each process)
        target_chunk = self.get_partial_data(
            data,
            total_proc_n=total_proc_n,
            current_proc_n=current_proc_n,
        )

        result: Any = func(target_chunk)

        return result

    def _get_total_proc_n(self, total_proc_n: Optional[int] = None) -> int:
        """Get total number of processes."""
        if total_proc_n is None:
            total_proc_n = self.total_proc_n
        assert (
            total_proc_n is not None
        ), f"Please set total_proc_n to the PartialProcess instance."
        return total_proc_n

    def _get_current_proc_n(self, current_proc_n: Optional[int] = None) -> int:
        """Get current process number."""
        if current_proc_n is None:
            current_proc_n = self.current_proc_n
        assert (
            current_proc_n is not None
        ), f"Please set current_proc_n to the PartialProcess instance."
        return current_proc_n

    def _get_func(self, func: Optional[Callable] = None) -> Callable:
        """Get function to process the data."""
        if func is None:
            func = self.func
        assert func is not None, f"Please set function to the PartialProcess instance."
        return func

    def get_partial_data(
        self,
        data: Union[List[Any], Dict],
        total_proc_n: Optional[int] = None,
        current_proc_n: Optional[int] = None,
    ) -> Union[List[Any], Dict]:
        # Get total number of processes and current process number
        total_proc_n = self._get_total_proc_n(total_proc_n)
        current_proc_n = self._get_current_proc_n(current_proc_n)

        # Divide the data into chunks (each chunk will be processed by each process)
        chunks: List = divide_into_chunks(data, num_chunks=total_proc_n)

        return chunks[current_proc_n]

    def divide_into_chunks(
        self,
        data: Union[List[Any], Dict],
        total_proc_n: Optional[int] = None,
    ) -> Union[List[Any], Dict]:
        total_proc_n = self._get_total_proc_n(total_proc_n)
        return divide_into_chunks(data, num_chunks=total_proc_n)

    def merge(
        self,
        results: List[Any],
        total_proc_n: Optional[int] = None,
    ) -> List[Any]:
        """Combine results from each process into a list with correct order."""

        # Get total number of processes and current process number
        total_proc_n = self._get_total_proc_n(total_proc_n)

        # Merge the results
        if isinstance(results, list):
            assert type(results[0]) == list, f"results must be list of list type."
            return do_flatten_list(results)
        elif isinstance(results, dict):
            assert (
                type(results[0]) == list
            ), f"value stored in the results dictionary must be list type."
            return do_flatten_list([results[i] for i in range(total_proc_n)])
        raise ValueError(
            f"results must be list or dict type, but {type(results)} is given."
        )


if __name__ == "__main__":
    pass



import unittest

import src.hkkang_utils.list as list_utils


class Test_list_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_list_utils, self).__init__(*args, **kwargs)

    def test_do_flatten_list(self):
        input_list = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        gold = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        result = list_utils.do_flatten_list(input_list)
        self.assertEqual(result, gold, f"result: {result}, gold_after: {gold}")

    def test_map_many(self):
        input_list = [1, 2, 3, 4, 5]
        functions = [lambda x: x + 1, lambda x: x * 2]
        gold = [4, 6, 8, 10, 12]
        result = list_utils.map_many(functions, input_list)
        self.assertEqual(result, gold, f"result: {result}, gold: {gold}")

    def test_get(self):
        input_list = [1, 2, 3, 4, 5]
        result1 = list_utils.get(input_list, 3, 0)
        gold1 = 4
        result2 = list_utils.get(input_list, 10, 0)
        gold2 = 0
        result3 = list_utils.get(input_list, -1, 0)
        gold3 = 0
        self.assertEqual(result1, gold1, f"result1: {result1}, gold1: {gold1}")
        self.assertEqual(result2, gold2, f"result2: {result2}, gold2: {gold2}")
        self.assertEqual(result3, gold3, f"result3: {result3}, gold3: {gold3}")

    def test_divide_into_chunks(self):
        item1 = list(range(10))
        pred = list_utils.divide_into_chunks(item1, 4)
        gold1 = [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

        # Validate results
        self.assertEqual(pred, gold1, f"pred: {pred}, gold1: {gold1}")

    def test_chunks(self):
        item1 = list(range(10))
        pred = list(list_utils.chunks(item1, 3))
        gold1 = [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

        # Validate results
        self.assertEqual(pred, gold1, f"pred: {pred}, gold1: {gold1}")


if __name__ == "__main__":
    unittest.main()



import abc
import unittest

import hkkang_utils.pattern as pattern_utils


class Test_pattern_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_pattern_utils, self).__init__(*args, **kwargs)

    def test_singleton_design(self):
        class SingletonClass(metaclass=pattern_utils.SingletonMeta):
            def __init__(self):
                self.a = 1

        a = SingletonClass()
        b = SingletonClass()
        self.assertEqual(a, b)

    def test_singleton_decorator(self):
        @pattern_utils.singleton
        class SingletonClass:
            def __init__(self):
                self.a = 1

        a = SingletonClass()
        b = SingletonClass()
        self.assertEqual(a, b)

    def test_signleton_with_args(self):
        class CustomClassWithArgs(metaclass=pattern_utils.SingletonMetaWithArgs):
            def __init__(self, a):
                self.a = a

        a = CustomClassWithArgs(1)
        b = CustomClassWithArgs(1)
        self.assertEqual(a, b)

        c = CustomClassWithArgs(2)
        self.assertNotEqual(a, c)

    def test_singleton_abc(self):
        class CustomAbstractClass(metaclass=pattern_utils.SingletonABCMeta):
            def __init__(self):
                self.a = 1

            @abc.abstractmethod
            def run(self):
                pass

        class CustomClass(CustomAbstractClass):
            def __init__(self):
                super().__init__()
                self.b = 2

            def run(self):
                pass

        a = CustomClass()
        b = CustomClass()
        self.assertEqual(a, b)

    def test_singleton_abc_with_args(self):
        class CustomAbstractClassWithArgs(
            metaclass=pattern_utils.SingletonABCMetaWithArgs
        ):
            def __init__(self, a):
                self.a = a

            @abc.abstractmethod
            def run(self):
                pass

        class CustomClassWithArgs(CustomAbstractClassWithArgs):
            def __init__(self, a, b):
                super().__init__(a)
                self.b = b

            def run(self):
                pass

        a = CustomClassWithArgs(1, 2)
        b = CustomClassWithArgs(1, 2)
        self.assertEqual(a, b, "a and b are not equal")

        c = CustomClassWithArgs(2, 3)
        self.assertNotEqual(a, c, "a and c are equal")


if __name__ == "__main__":
    unittest.main()



import logging


def init(format: str = None, datefmt: str = None, level: int = None) -> None:
    if format is None:
        format = "[%(asctime)s %(levelname)s %(name)s] %(message)s"
    if datefmt is None:
        datefmt = "%m/%d %H:%M:%S"
    if level is None:
        level = logging.INFO

    # Set logging format
    logging.basicConfig(
        format=format,
        datefmt=datefmt,
        level=level,
    )



import math
from typing import Type, Union


class Training_context:
    """Syntatic sugar to keep track of training step"""

    def __init__(
        self,
        Trainer: Type,
        effective_batch_size: int,
        dataloader_batch_size: int,
        eval_freq_estep: int,
        max_estep: int,
    ):
        self.Trainer = Trainer
        self.effective_batch_size = effective_batch_size
        self.dataloader_batch_size = dataloader_batch_size
        self.eval_freq_estep = eval_freq_estep
        self.max_estep = max_estep
        self._steps_per_estep: Union[int, None] = None

    def __enter__(self):
        self.Trainer.step += 1
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    @property
    def steps_per_estep(self):
        if not self._steps_per_estep:
            self._steps_per_estep = math.ceil(
                self.effective_batch_size / self.dataloader_batch_size
            )
        return self._steps_per_estep

    @property
    def step(self):
        return self.Trainer.step

    @property
    def estep(self):
        return self.Trainer.step // self.steps_per_estep

    @property
    def is_state_to_update(self):
        return self.step != 0 and not (self.step % self.steps_per_estep)

    @property
    def is_state_to_eval(self):
        return self.is_state_to_update and not (self.estep % self.eval_freq_estep)

    @property
    def is_state_to_exit(self):
        return self.estep >= self.max_estep


if __name__ == "__main__":
    pass
